# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project

"""
Shared code used by tools for building the Ansible distribution
"""

from __future__ import annotations

__version__ = "3.5.0"

__all__ = ("__version__",)
